# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, Dict, List

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.get_quantizer_config import \
    get_trainable_quantizer_weights_config, get_trainable_quantizer_activation_config, \
    get_trainable_quantizer_quantization_candidates
from model_compression_toolkit.qat.keras.quantizer.base_keras_qat_quantizer import BaseKerasQATTrainableQuantizer
from model_compression_toolkit.qat.common.qat_config import QATConfig
from model_compression_toolkit.quantizers_infrastructure import QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.trainable_infrastructure.common.get_quantizers import \
    get_trainable_quantizer_class


def quantization_builder(n: common.BaseNode,
                         qat_config: QATConfig,
                         fw_info: FrameworkInfo,
                         ) -> Tuple[Dict[str, BaseKerasQATTrainableQuantizer], List[BaseKerasQATTrainableQuantizer]]:
    """
    Build quantizers for a node according to its quantization configuration.

    Args:
        n: Node to build its QuantizeConfig.
        qat_config (QATConfig): QAT configuration
        fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).

    Returns:
        weights_quantizers: A dictionary between a weight's name to its quantizer.
        activation_quantizers: A list of activations quantization, one for each layer output.
    """
    if len(n.candidates_quantization_cfg) > 1:
        wq_cand, aq_cand = get_trainable_quantizer_quantization_candidates(n)
    else:
        wq_cand, aq_cand = None, None

    weight_quantizers = {}
    if n.is_weights_quantization_enabled():
        quant_method = n.final_weights_quantization_cfg.weights_quantization_method

        quantizer_class = get_trainable_quantizer_class(QuantizationTarget.Weights,
                                                        qat_config.weight_training_method,
                                                        quant_method,
                                                        BaseKerasQATTrainableQuantizer)
        attributes = fw_info.get_kernel_op_attributes(n.type)
        for attr in attributes:
            weight_quantizers.update({attr: quantizer_class(get_trainable_quantizer_weights_config(n, wq_cand),
                                                            **qat_config.weight_quantizer_params_override)})

    activation_quantizers = []
    if n.is_activation_quantization_enabled():
        quant_method = n.final_activation_quantization_cfg.activation_quantization_method
        # single output -> normalize to list of output_shapes
        output_shapes = n.output_shape if isinstance(n.output_shape[0], (list, tuple)) else [n.output_shape]

        quantizer_class = get_trainable_quantizer_class(QuantizationTarget.Activation,
                                                        qat_config.activation_training_method,
                                                        quant_method,
                                                        BaseKerasQATTrainableQuantizer)

        activation_quantizers = [quantizer_class(get_trainable_quantizer_activation_config(n, aq_cand),
                                                 **qat_config.activation_quantizer_params_override)] * len(output_shapes)

    return weight_quantizers, activation_quantizers



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf
from typing import Tuple, List

from model_compression_toolkit.core.keras.constants import USE_BIAS
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from tensorflow.keras.models import Model

from model_compression_toolkit.core.keras.default_framework_info import DEFAULT_KERAS_INFO
from model_compression_toolkit.gptq.common.gptq_graph import get_kernel_attribute_name_for_gptq
from model_compression_toolkit.quantizers_infrastructure import KerasQuantizationWrapper


def get_gptq_trainable_parameters(fxp_model: Model,
                                  fw_info: FrameworkInfo,
                                  add_bias: bool = False) -> (
        List[tf.Variable], List[tf.Variable], List[tf.Variable], List[tf.Variable], List[tf.Variable]):
    """
    Get trainable parameters from all layers in a model

    Args:
        fxp_model: Model to get its trainable parameters.
        fw_info: Framework information needed for keras kernel ops list.
        add_bias: Whether to include biases of the model (if there are) or not.

    Returns:
        A list of trainable variables in a model. Each item is a list of a layers weights.
    """

    trainable_weights: List[tf.Tensor] = []
    trainable_threshold: List[tf.Tensor] = []
    bias_weights: List[List[tf.Tensor]] = []
    temperature_weights: List[tf.Tensor] = []

    for layer in fxp_model.layers:
        if isinstance(layer, KerasQuantizationWrapper):
            kernel_attribute = get_kernel_attribute_name_for_gptq(layer_type=type(layer.layer),
                                                                  fw_info=DEFAULT_KERAS_INFO)

            # collect trainable weights per layer
            layer_trainable_weights = layer.weights_quantizers[kernel_attribute].get_aux_variable()
            layer_trainable_threshold = layer.weights_quantizers[kernel_attribute].get_quantization_variable()

            if add_bias:
                kernel_ops_attrs = fw_info.kernel_ops_attributes_mapping.get(type(layer.layer))
                use_bias = kernel_ops_attrs is not None and kernel_ops_attrs[0] is not None \
                           and layer.layer.get_config().get(USE_BIAS)
                if use_bias is not None and use_bias:
                    bias_weights.append([layer.layer.bias])
            trainable_weights.append(layer_trainable_weights)
            trainable_threshold.extend(layer_trainable_threshold)

    return trainable_weights, bias_weights, trainable_threshold, temperature_weights


def get_weights_for_loss(fxp_model: Model) -> Tuple[List[list], List[list]]:
    """
    Get all float and quantized kernels for the GPTQ loss

    Args:
        fxp_model: Model to get its float and quantized weights.

    Returns:
        A list of float kernels, each item is the float kernel of the layer
        A list of quantized kernels, each item is the quantized kernel of the layer
    """

    flp_weights_list = []
    fxp_weights_list = []
    for layer in fxp_model.layers:
        if isinstance(layer, KerasQuantizationWrapper):

            # collect pairs of float and quantized weights per layer
            _layer_flp_weights, _layer_fxp_weights = [], []
            for weight, quantizer_vars, quantizer in layer.get_weights_vars():
                _layer_flp_weights.append(quantizer_vars)
                _layer_fxp_weights.append(quantizer(training=False, inputs=quantizer_vars))

            flp_weights_list.append(_layer_flp_weights)
            fxp_weights_list.append(_layer_fxp_weights)

    return flp_weights_list, fxp_weights_list


# TODO: this function need to move to location that is relevant only for soft quantizer -
#  once deciding how to handle GPTQ quantizers regularization.
def get_soft_rounding_reg(fxp_model: Model) -> List[tf.Tensor]:
    """
    This function returns the soft quantizer regularization values for SoftRounding.

    Args:
        fxp_model: A model to be quantized with SoftRounding.

    Returns: A list of tensors.
    """

    soft_reg_aux: List[tf.Tensor] = []
    for layer in fxp_model.layers:
        if isinstance(layer, KerasQuantizationWrapper):
            kernel_attribute = get_kernel_attribute_name_for_gptq(layer_type=type(layer.layer),
                                                                  fw_info=DEFAULT_KERAS_INFO)

            soft_reg_aux.append(layer.weights_quantizers[kernel_attribute].get_regularization())
    return soft_reg_aux



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple
import torch


def ste_round(x: torch.Tensor) -> torch.Tensor:
    """
    Calculate the rounded values of a tensor
    Args:
        x: input variable
    Returns:
        rounded value
    """
    return (torch.round(x) - x).detach() + x


def ste_clip(x: torch.Tensor, min_val=-1.0, max_val=1.0) -> torch.Tensor:
    """
    Clip a variable between fixed values such that min_val<=output<=max_val
    Args:
        x: input variable
        min_val: minimum value for clipping
        max_val: maximum value for clipping
    Returns:
        clipped variable
    """
    return (torch.clip(x, min=min_val, max=max_val) - x).detach() + x


def fix_range_to_include_zero(range_min: torch.Tensor,
                              range_max: torch.Tensor,
                              n_bits: int) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Adjusting the quantization range to include representation of 0.0 in the quantization grid.
    If quantization per-channel, then range_min and range_max should be tensors in the specific shape that allows
    quantization along the channel_axis.
    Args:
        range_min: min bound of the quantization range (before adjustment).
        range_max: max bound of the quantization range (before adjustment).
        n_bits: Number of bits to quantize the tensor.
    Returns: adjusted quantization range
    """
    min_positive = range_min > 0
    max_negative = range_max < 0
    mid_range = torch.logical_and(torch.logical_not(min_positive), torch.logical_not(max_negative))
    min_positive = min_positive.float()
    max_negative = max_negative.float()
    mid_range = mid_range.float()

    scale = (range_max - range_min) / (2 ** n_bits - 1)
    min_range_adj = scale * torch.round(range_min / scale)
    max_range_adj = range_max - range_min + min_range_adj

    min_range_adj = min_range_adj * mid_range + max_negative * range_min
    max_range_adj = max_range_adj * mid_range + min_positive * range_max
    return min_range_adj, max_range_adj


def symmetric_quantizer(tensor_data: torch.Tensor,
                        threshold: torch.Tensor,
                        n_bits: int,
                        sign: bool = False) -> torch.Tensor:
    """
    Quantize a tensor according to the number of bits and threshold.
    Symmetric quantization.
    Args:
        tensor_data: Tensor values to quantize.
        threshold: threshold for quantization.
        n_bits: Number of bits to quantize the tensor.
        sign: sign of tensor_data
    Returns:
        Quantized data.
    """

    # Compute the step size of quantized values.
    n_pos = 2 ** (n_bits - int(sign))
    delta_tensor = threshold / n_pos

    # Compute min/max int value
    min_val = -int(sign) * n_pos
    max_val = n_pos - 1

    # Apply rounding
    input_tensor_int = ste_round(tensor_data / delta_tensor)

    # Clip data in range
    clipped_tensor = ste_clip(input_tensor_int, min_val=min_val, max_val=max_val)

    # Quantize the data between -threshold/threshold
    q = delta_tensor * clipped_tensor
    return q


def uniform_quantizer(tensor_data: torch.Tensor,
                      range_min: torch.Tensor,
                      range_max: torch.Tensor,
                      n_bits: int) -> torch.Tensor:
    """
    Quantize a tensor according to given range (min, max) and number of bits.
    Uniform quantization.
    Args:
        tensor_data: Tensor values to quantize.
        range_min: minimum bound of the range for quantization (or array of min values per channel).
        range_max: maximum bound of the range for quantization (or array of max values per channel).
        n_bits: Number of bits to quantize the tensor.
    Returns:
        Quantized data.
    """
    # adjusts the quantization range so the quantization grid includes zero.
    a, b = fix_range_to_include_zero(range_min, range_max, n_bits)

    # Compute the step size of quantized values.
    delta_tensor = (b - a) / (2 ** n_bits - 1)

    # Apply rounding
    input_tensor_int = ste_round((tensor_data - a) / delta_tensor)

    # Clip data in range
    clipped_tensor = ste_clip(input_tensor_int, min_val=0, max_val=2 ** n_bits - 1)

    # Quantize the data between min/max of quantization range.
    q = delta_tensor * clipped_tensor + a
    return q



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import warnings
from typing import List

import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TF
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer, \
    QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.constants import MULTIPLIER_N_BITS, EPS

if FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.base_keras_inferable_quantizer import \
        BaseKerasInferableQuantizer
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizer_utils import \
        lut_quantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.LUT_SYM_QUANTIZER],
                    quantizer_type=None)
    class WeightsLUTSymmetricInferableQuantizer(BaseKerasInferableQuantizer):
        """
        Class for quantizing weights using a lut symmetric quantizer
        """

        def __init__(self,
                     num_bits: int,
                     cluster_centers: np.ndarray,
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     input_rank: int = None,
                     multiplier_n_bits: int = MULTIPLIER_N_BITS,
                     eps: float = EPS):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                cluster_centers: the cluster centers to assign the weights
                threshold: threshold for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: axis along which to apply per-channel quantization
                input_rank: number of dimensions of input tensor the quantizer quantizes
                multiplier_n_bits: Number of bits that determines the quantization range
                eps: Small value for numerical stability in division
            """

            super(WeightsLUTSymmetricInferableQuantizer, self).__init__()

            assert isinstance(threshold, list), f'Expected threshold to be of type list but is {type(threshold)}'
            assert all([isinstance(x, (float, np.float32, np.float64)) for x in
                        threshold]), f'Expected threshold list to contain float or np.float values but found ' \
                                     f'{[type(x) for x in threshold]}'

            self.threshold = np.asarray(threshold)

            if per_channel:
                assert input_rank is not None, f'Input rank is missing in per channel quantization'
                assert channel_axis is not None, f'Channel axis is missing in per channel quantization'
                assert len(threshold) >= 1, f'In per-channel quantization threshold list should be of length >= 1 ' \
                                            f'but is {len(threshold)} '
            else:
                assert len(threshold) == 1, f'In per-tensor quantization threshold should be of length 1 but is' \
                                            f' {len(threshold)}'
                self.threshold = self.threshold[0]

            assert len(np.unique(cluster_centers)) <= 2 ** num_bits, \
                f'Expected num of cluster centers to be less or equal than {2 ** num_bits} ' \
                f'but got {len(cluster_centers)}'

            assert not np.any(cluster_centers - cluster_centers.astype(int)), f'Expected cluster centers to be integers'

            # Weight quantization is signed, hence the quantization range is
            # [-2**(multiplier_n_bits - 1), 2**(multiplier_n_bits - 1) - 1]
            assert np.all((-1 * (2 ** (multiplier_n_bits - 1)) <= cluster_centers) &
                          (cluster_centers <= (2 ** (multiplier_n_bits - 1) - 1))), \
                f'Expected cluster centers in the quantization range'

            # num_bits must be less than multiplier_n_bits
            assert num_bits <= multiplier_n_bits, f'Look-Up-Table bit configuration has {num_bits} bits. It must be ' \
                                                  f'less then {multiplier_n_bits}'
            if num_bits == multiplier_n_bits:
                warnings.warn("Num of bits equal to multiplier n bits, Please be aware LUT quantizier may be "
                              "inefficient in that case, consider using SymmetricInferableQuantizer instead")

            self.num_bits = num_bits
            self.cluster_centers = cluster_centers
            self.multiplier_n_bits = multiplier_n_bits
            self.eps = eps
            self.per_channel = per_channel
            self.channel_axis = channel_axis
            self.input_rank = input_rank

            # Tensorflow's fake_quant_with_min_max_vars_per_channel only works on last axis, so
            # need to move the quantization axis to the last axis
            if per_channel and channel_axis not in [-1, self.input_rank - 1]:
                # If per-channel quantization is being used and the channel axis is not the last axis,
                # create a permutation vector to move the channel axis to the last position
                self.perm_vec = list(np.arange(self.input_rank))
                self.perm_vec[channel_axis] = self.input_rank - 1
                self.perm_vec[self.input_rank - 1] = channel_axis
            else:
                # If per-channel quantization is not being used or the channel axis is already the last axis,
                # set the permutation vector to None
                self.perm_vec = None

        def __call__(self, inputs: tf.Tensor) -> tf.Tensor:
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            assert inputs.dtype == tf.float32, f'Input tensor was expected to be a float tensor but is of type ' \
                                               f'{inputs.dtype}'

            # If per-channel quantization is being used
            if self.per_channel:
                # If a permutation vector has been created to move the channel axis to the last position
                if self.perm_vec:
                    # Transpose the input tensor to move the channel axis to the last position
                    inputs = tf.transpose(inputs, perm=self.perm_vec)

                # Quantize the input tensor using per-channel quantization
                q_tensor = lut_quantizer(inputs, cluster_centers=self.cluster_centers, signed=True,
                                         threshold=self.threshold, multiplier_n_bits=self.multiplier_n_bits,
                                         eps=self.eps)
                if self.perm_vec:
                    # Transpose the quantized tensor back to its original shape
                    q_tensor = tf.transpose(q_tensor, perm=self.perm_vec)

                # Return the quantized tensor
                return q_tensor
            else:
                return lut_quantizer(inputs, cluster_centers=self.cluster_centers, signed=True,
                                     threshold=self.threshold, multiplier_n_bits=self.multiplier_n_bits, eps=self.eps)

        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'per_channel', 'num_bits', 'cluster_centers', 'threshold',
                 'channel_axis', 'input_rank', 'multiplier_n_bits', 'eps'
            """
            return {'per_channel': self.per_channel,
                    'num_bits': self.num_bits,
                    'cluster_centers': self.cluster_centers,
                    'threshold': self.threshold,
                    'channel_axis': self.channel_axis,
                    'input_rank': self.input_rank,
                    'multiplier_n_bits': self.multiplier_n_bits,
                    'eps': self.eps}

else:
    class WeightsLUTSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing tensorflow and tensorflow_model_optimization is mandatory '
                            'when using WeightsLUTSymmetricInferableQuantizer. '
                            'Could not find Tensorflow package.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, Any

from tensorflow.keras.layers import BatchNormalization, DepthwiseConv2D, Conv2DTranspose, Conv2D

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher, \
    NodeFrameworkAttrMatcher
from model_compression_toolkit.core.common.substitutions.batchnorm_reconstruction import \
    BatchNormalizationReconstruction
from model_compression_toolkit.core.keras.constants import LINEAR, ACTIVATION, GAMMA, \
    BETA, MOVING_MEAN, MOVING_VARIANCE, EPSILON, EPSILON_VAL, MOMENTUM, MOMENTUM_VAL


def batchnorm_reconstruction_node_matchers() -> NodeOperationMatcher:
    """
    Match linear layers for BN reconstruction.

    Returns:
        Node matcher for convolution type nodes.
    """
    conv_node = NodeOperationMatcher(DepthwiseConv2D) | \
                NodeOperationMatcher(Conv2D) | \
                NodeOperationMatcher(Conv2DTranspose)

    activation_linear = NodeFrameworkAttrMatcher(ACTIVATION, LINEAR)
    source_node = conv_node & activation_linear
    return source_node


def create_bn_node(source_node: BaseNode,
                   bn_node_weights: Dict[Any, Any]):
    """
    Creates BN nodes after linear nodes.
    Args:
        source_node: Node matcher for convolution type nodes.
        bn_node_weights: dictionary of BN weights.
    """
    bn_node = BaseNode(name=source_node.name + '_reconstructed',
                       framework_attr={EPSILON: EPSILON_VAL, MOMENTUM: MOMENTUM_VAL},
                       input_shape=source_node.output_shape,
                       output_shape=source_node.output_shape,
                       weights=bn_node_weights,
                       layer_class=BatchNormalization)
    return bn_node


def keras_batchnorm_reconstruction() -> BatchNormalizationReconstruction:
    """
    Reconstruct BatchNormalization after linear layers.
    Returns:
        A BatchNormalizationReconstruction initialized for keras models.
    """
    source_node = batchnorm_reconstruction_node_matchers()
    return BatchNormalizationReconstruction(source_node,
                                            create_bn_node,
                                            GAMMA,
                                            BETA,
                                            MOVING_MEAN,
                                            MOVING_VARIANCE,
                                            EPSILON_VAL)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Dict, Union

import numpy as np
import tensorflow as tf
from tensorflow.python.framework.tensor_shape import TensorShape
from model_compression_toolkit.core.common.constants import SIGNED

from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.qat.common import THRESHOLD_TENSOR
from model_compression_toolkit.qat.common.constants import FQ_MIN, FQ_MAX
from model_compression_toolkit import quantizers_infrastructure as qi, TrainingMethod
from model_compression_toolkit.core.common import constants as C
from model_compression_toolkit.qat.keras.quantizer.base_keras_qat_quantizer import BaseKerasQATTrainableQuantizer
from model_compression_toolkit.quantizers_infrastructure import TrainableQuantizerWeightsConfig, \
    TrainableQuantizerActivationConfig
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers import \
    WeightsPOTInferableQuantizer, WeightsSymmetricInferableQuantizer, ActivationPOTInferableQuantizer, \
    ActivationSymmetricInferableQuantizer


@mark_quantizer(quantization_target=qi.QuantizationTarget.Weights,
                quantization_method=[QuantizationMethod.POWER_OF_TWO, QuantizationMethod.SYMMETRIC],
                quantizer_type=TrainingMethod.STE)
class STEWeightQATQuantizer(BaseKerasQATTrainableQuantizer):
    """
    Trainable constrained quantizer to quantize a layer inputs.
    """

    def __init__(self, quantization_config: TrainableQuantizerWeightsConfig):
        """
        Initialize a TrainableWeightQuantizer object with parameters to use
        for the quantization.

        Args:
            quantization_config: trainable quantizer config class
        """
        super().__init__(quantization_config)
        self.power_of_two = quantization_config.weights_quantization_method == QuantizationMethod.POWER_OF_TWO
        self.threshold_values = quantization_config.weights_quantization_params[C.THRESHOLD]
        self.threshold_shape = np.asarray(self.threshold_values).shape
        self.per_channel = self.quantization_config.weights_per_channel_threshold
        self.channel_axis = self.quantization_config.weights_channels_axis
        self.np_threshold_values = np.reshape(np.asarray(self.threshold_values),[-1]) if self.channel_axis else float(self.threshold_values)

        if self.per_channel and self.channel_axis not in [-1, len(self.threshold_shape) - 1]:
            # Tensorflow's fake_quant_with_min_max_vars_per_channel only works on last axis, so
            # need to move the quantization axis to the last axis
            self.perm_vec = list(np.arange(len(self.threshold_shape)))
            self.perm_vec[self.channel_axis] = len(self.threshold_shape) - 1
            self.perm_vec[len(self.threshold_shape) - 1] = self.channel_axis
        else:
            self.perm_vec = None

        if self.power_of_two:
            self.np_threshold_values = np.power(2.0,np.ceil(np.log2(np.maximum(self.np_threshold_values, C.MIN_THRESHOLD))))

        self.num_bits = self.quantization_config.weights_n_bits
        delta = self.np_threshold_values / np.power(2.0, self.num_bits - int(C.WEIGHTS_SIGNED))
        min_int = -int(C.WEIGHTS_SIGNED) * (2 ** (self.num_bits - int(C.WEIGHTS_SIGNED)))
        max_int = (2 ** (self.num_bits - int(C.WEIGHTS_SIGNED))) - 1
        self.min = delta * min_int
        self.max = delta * max_int
        self.quantizer_parameters = {}

    def initialize_quantization(self,
                                tensor_shape: TensorShape,
                                name: str,
                                layer: qi.KerasQuantizationWrapper) -> Dict[str, tf.Variable]:
        """
        Add min and max variables to layer.
        Args:
            tensor_shape: Tensor shape the quantizer quantize.
            name: Prefix of variables names.
            layer: Layer to add the variables to. The variables are saved
            in the layer's scope.

        Returns:
            Dictionary of new variables.
        """
        ptq_threshold_tensor = layer.add_weight(
            name + THRESHOLD_TENSOR,
            shape=len(self.np_threshold_values) if self.channel_axis else (),
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False)
        ptq_threshold_tensor.assign(self.np_threshold_values)

        fq_min = layer.add_weight(
            name + FQ_MIN,
            shape=len(self.min) if self.channel_axis else (),
            initializer=tf.keras.initializers.Constant(-1.0),
            trainable=False)
        fq_min.assign(self.min)

        fq_max = layer.add_weight(
            name + FQ_MAX,
            shape=len(self.max) if self.channel_axis else (),
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False)
        fq_max.assign(self.max)

        # save the quantizer added parameters for later calculations
        self.quantizer_parameters = {THRESHOLD_TENSOR: ptq_threshold_tensor,
                                     FQ_MIN: fq_min, FQ_MAX: fq_max}
        return self.quantizer_parameters

    def __call__(self,
                 inputs: tf.Tensor,
                 training: bool):
        """
        Quantize a tensor.
        Args:
            inputs: Input tensor to quantize.
            training: Whether the graph is in training mode.
            weights: Dictionary of weights the quantizer can use to quantize the tensor.
            **kwargs: Additional variables the quantizer may receive.

        Returns:
            The quantized tensor.
        """

        _min = self.quantizer_parameters[FQ_MIN]
        _max = self.quantizer_parameters[FQ_MAX]
        if self.channel_axis:
            if self.perm_vec:
                inputs = tf.transpose(inputs, perm=self.perm_vec)
            q_tensor = tf.quantization.fake_quant_with_min_max_vars_per_channel(inputs, _min, _max,
                                                                                num_bits=self.num_bits)
            if self.perm_vec:
                q_tensor = tf.transpose(q_tensor, perm=self.perm_vec)
        else:
            q_tensor = tf.quantization.fake_quant_with_min_max_vars(inputs, _min, _max,
                                                                    num_bits=self.num_bits)

        return q_tensor

    def convert2inferable(self) -> Union[WeightsPOTInferableQuantizer, WeightsSymmetricInferableQuantizer]:
        """
        Convert quantizer to inferable quantizer.

        Returns:
            BaseKerasInferableQuantizer object.
        """
        if self.power_of_two:
            pot_threshold = 2 ** np.ceil(np.log2(self.quantizer_parameters[THRESHOLD_TENSOR]))
            return WeightsPOTInferableQuantizer(num_bits=self.num_bits,
                                                threshold=list(pot_threshold.flatten()),
                                                per_channel=self.per_channel,
                                                channel_axis=self.channel_axis,
                                                input_rank=len(self.threshold_shape))
        else:
            return WeightsSymmetricInferableQuantizer(num_bits=self.num_bits,
                                                      threshold=list(self.quantizer_parameters[
                                                                         THRESHOLD_TENSOR].numpy().flatten()),
                                                      per_channel=self.per_channel,
                                                      channel_axis=self.channel_axis,
                                                      input_rank=len(self.threshold_shape))


@mark_quantizer(quantization_target=qi.QuantizationTarget.Activation,
                quantization_method=[QuantizationMethod.POWER_OF_TWO, QuantizationMethod.SYMMETRIC],
                quantizer_type=TrainingMethod.STE)
class STEActivationQATQuantizer(BaseKerasQATTrainableQuantizer):
    """
    Trainable constrained quantizer to quantize a layer outputs.
    """

    def __init__(self, quantization_config: TrainableQuantizerActivationConfig):
        """
        Initialize a STEActivationQATQuantizer object with parameters to use
        for the quantization.

        Args:
            quantization_config: trainable quantizer config class
        """
        super().__init__(quantization_config)
        self.power_of_two = quantization_config.activation_quantization_method == QuantizationMethod.POWER_OF_TWO
        self.threshold_values = quantization_config.activation_quantization_params[C.THRESHOLD]
        self.threshold_shape = np.asarray(self.threshold_values).shape
        self.np_threshold_values = float(self.threshold_values)
        self.signed = quantization_config.activation_quantization_params[SIGNED]
        if self.power_of_two:
            self.np_threshold_values = np.power(2.0,
                                                np.ceil(np.log2(np.maximum(self.np_threshold_values, C.MIN_THRESHOLD))))
        self.num_bits = quantization_config.activation_n_bits
        delta = self.np_threshold_values / np.power(2.0, self.num_bits - int(self.signed))
        min_int = -int(self.signed) * (2 ** (self.num_bits - int(self.signed)))
        max_int = (2 ** (self.num_bits - int(self.signed))) - 1
        self.min = delta * min_int
        self.max = delta * max_int
        self.quantizer_parameters = {}

    def initialize_quantization(self,
                                tensor_shape: TensorShape,
                                name: str,
                                layer: qi.KerasQuantizationWrapper) -> Dict[str, tf.Variable]:
        """
        Add min and max variables to layer.
        Args:
            tensor_shape: Tensor shape the quantizer quantize.
            name: Prefix of variables names.
            layer: Layer to add the variables to. The variables are saved
            in the layer's scope.

        Returns:
            Dictionary of new variables.
        """
        ptq_threshold_tensor = layer.add_weight(
            name + THRESHOLD_TENSOR,
            shape=(),
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False)
        ptq_threshold_tensor.assign(self.np_threshold_values)

        fq_min = layer.add_weight(
            name + FQ_MIN,
            shape=(),
            initializer=tf.keras.initializers.Constant(-1.0),
            trainable=False)
        fq_min.assign(self.min)

        fq_max = layer.add_weight(
            name + FQ_MAX,
            shape=(),
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False)
        fq_max.assign(self.max)

        # save the quantizer added parameters for later calculations
        self.quantizer_parameters = {THRESHOLD_TENSOR: ptq_threshold_tensor,
                                     FQ_MIN: fq_min, FQ_MAX: fq_max}
        return self.quantizer_parameters

    def __call__(self,
                 inputs: tf.Tensor,
                 training: bool):
        """
        Quantize a tensor.
        Args:
            inputs: Input tensor to quantize.
            training: Whether the graph is in training mode.

        Returns:
            The quantized tensor.
        """

        _min = self.quantizer_parameters[FQ_MIN]
        _max = self.quantizer_parameters[FQ_MAX]
        q_tensor = tf.quantization.fake_quant_with_min_max_vars(inputs, _min, _max,
                                                                num_bits=self.num_bits)

        return q_tensor

    def convert2inferable(self) -> Union[ActivationPOTInferableQuantizer, ActivationSymmetricInferableQuantizer]:
        """
        Convert quantizer to inferable quantizer.

        Returns:
            BaseKerasInferableQuantizer object.
        """

        if self.power_of_two:
            pot_threshold = 2 ** np.ceil(np.log2(self.quantizer_parameters[THRESHOLD_TENSOR]))
            return ActivationPOTInferableQuantizer(num_bits=self.num_bits,
                                                      # In activation quantization is per-tensor only - thus we pass
                                                      # the threshold as a list with a len of 1
                                                      threshold=[pot_threshold],
                                                      signed=self.signed)
        else:
            return ActivationSymmetricInferableQuantizer(num_bits=self.num_bits,
                                                         # In activation quantization is per-tensor only - thus we
                                                         # pass the threshold as a list with a len of 1
                                                         threshold=[
                                                             self.quantizer_parameters[THRESHOLD_TENSOR].numpy()],
                                                         signed=self.signed)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Any

from model_compression_toolkit.core.common.target_platform.operators import OperatorSetConcat
from model_compression_toolkit.core.common.target_platform.target_platform_model_component import TargetPlatformModelComponent


class Fusing(TargetPlatformModelComponent):

    def __init__(self, operator_groups_list, name=None):
        assert isinstance(operator_groups_list,
                          list), f'List of operator groups should be of type list but is {type(operator_groups_list)}'
        assert len(operator_groups_list) >= 2, f'Fusing can not be created for a single operators group'
        if name is None:
            name = '_'.join([x.name for x in operator_groups_list])
        super().__init__(name)
        self.operator_groups_list = operator_groups_list

    def contains(self, other: Any):
        if not isinstance(other, Fusing):
            return False
        for i in range(len(self.operator_groups_list) - len(other.operator_groups_list) + 1):
            for j in range(len(other.operator_groups_list)):
                if self.operator_groups_list[i + j] != other.operator_groups_list[j] and not (isinstance(self.operator_groups_list[i + j], OperatorSetConcat) and (other.operator_groups_list[j] in self.operator_groups_list[i + j].op_set_list)):
                    break
            else:
                return True
        return False


    def get_info(self):
        if self.name is not None:
            return {self.name: ' -> '.join([x.name for x in self.operator_groups_list])}
        return ' -> '.join([x.name for x in self.operator_groups_list])




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import tensorflow as tf
from tensorflow.keras.layers import Layer

from model_compression_toolkit import quantizers_infrastructure as qi
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Graph
from model_compression_toolkit.core.common.user_info import UserInformation
from model_compression_toolkit.core.keras.back2framework.keras_model_builder import KerasModelBuilder
from model_compression_toolkit.exporter.model_wrapper.keras.builder.node_to_quantizers import \
    get_quantization_quantizers


def _get_wrapper(node: common.BaseNode,
                 layer: Layer) -> qi.KerasQuantizationWrapper:
    """
    A function which takes a computational graph node and a keras layer and perform the quantization wrapping
    Args:
        n: A node of mct graph.
        layer: A keras layer

    Returns: Wrapped layer with weights quantizers and activation quantizers

    """
    weights_quantizers, activation_quantizers = get_quantization_quantizers(node)
    return qi.KerasQuantizationWrapper(layer, weights_quantizers, activation_quantizers)


def get_exportable_keras_model(graph: Graph) -> Tuple[tf.keras.models.Model,UserInformation]:
    """
    Convert graph to an exportable Keras model (model with all quantization parameters).
    An exportable model can then be exported using model_exporter, to retrieve the
    final exported model.

    Args:
        graph: Graph to convert to an exportable Keras model.

    Returns:
        Exportable Keras model and user information.
    """
    exportable_model, user_info = KerasModelBuilder(graph=graph,
                                         wrapper=_get_wrapper).build_model()
    exportable_model.trainable = False
    return exportable_model, user_info



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

class WrapperQuantizeConfig:
    """
    Configuration for how a layer should be quantized.
    """

    def __init__(self,
                 is_weight_quantized: bool,
                 is_activation_quantized: bool
                 ):
        """

        Args:
            is_weight_quantized: Whether weights quantization is enabled or not.
            is_activation_quantized: Whether activation quantization is enabled or not.
        """

        self.is_weight_quantized = is_weight_quantized
        self.is_activation_quantized = is_activation_quantized

    def get_weight_quantizers(self) -> list:
        """

        Returns: A List of quantizers for weights quantization.

        """
        raise NotImplemented  # pragma: no cover

    def get_activation_quantizers(self) -> list:
        """

        Returns: A List of quantizers for activation quantization.

        """
        raise NotImplemented  # pragma: no cover






# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.constants import FOUND_TORCH
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import PYTORCH
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfigV2, RoundingType
from model_compression_toolkit.core.common.target_platform import TargetPlatformCapabilities
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit.core.runner import core_runner, _init_tensorboard_writer
from model_compression_toolkit.gptq.runner import gptq_runner
from model_compression_toolkit.core.exporter import export_model
from model_compression_toolkit.core.analyzer import analyzer_model_quantization
from model_compression_toolkit import CoreConfig, GPTQQuantizerConfig
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfigV2

LR_DEFAULT = 1e-4
LR_REST_DEFAULT = 1e-4
LR_BIAS_DEFAULT = 1e-4
LR_QUANTIZATION_PARAM_DEFAULT = 1e-4

if FOUND_TORCH:
    from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
    from model_compression_toolkit.core.pytorch.pytorch_implementation import PytorchImplementation
    from model_compression_toolkit.core.pytorch.constants import DEFAULT_TP_MODEL
    from model_compression_toolkit.gptq.pytorch.gptq_loss import multiple_tensors_mse_loss
    from model_compression_toolkit.exporter.model_wrapper.pytorch.builder.fully_quantized_model_builder import get_exportable_pytorch_model
    import torch
    from torch.nn import Module
    from torch.optim import Adam, Optimizer
    from model_compression_toolkit import get_target_platform_capabilities


    DEFAULT_PYTORCH_TPC = get_target_platform_capabilities(PYTORCH, DEFAULT_TP_MODEL)


    def get_pytorch_gptq_config(n_epochs: int,
                                optimizer: Optimizer = Adam([torch.Tensor([])], lr=LR_DEFAULT),
                                optimizer_rest: Optimizer = Adam([torch.Tensor([])], lr=LR_REST_DEFAULT),
                                loss: Callable = multiple_tensors_mse_loss,
                                log_function: Callable = None) -> GradientPTQConfigV2:
        """
        Create a GradientPTQConfigV2 instance for Pytorch models.

        args:
            n_epochs (int): Number of epochs for running the representative dataset for fine-tuning.
            optimizer (Optimizer): Pytorch optimizer to use for fine-tuning for auxiliry variable.
            optimizer_rest (Optimizer): Pytorch optimizer to use for fine-tuning of the bias variable.
            loss (Callable): loss to use during fine-tuning. should accept 4 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors, the 3rd is a list of quantized weights and the 4th is a list of float weights.
            log_function (Callable): Function to log information about the gptq process.

        returns:
            a GradientPTQConfigV2 object to use when fine-tuning the quantized model using gptq.

        Examples:

            Import MCT and Create a GradientPTQConfigV2 to run for 5 epochs:

            >>> import model_compression_toolkit as mct
            >>> gptq_conf = mct.get_pytorch_gptq_config(n_epochs=5)

            Other PyTorch optimizers can be passed with dummy params:

            >>> import torch
            >>> gptq_conf = mct.get_pytorch_gptq_config(n_epochs=3, optimizer=torch.optim.Adam([torch.Tensor(1)]))

            The configuration can be passed to :func:`~model_compression_toolkit.pytorch_post_training_quantization` in order to quantize a pytorch model using gptq.

        """
        bias_optimizer = Adam([torch.Tensor([])], lr=LR_BIAS_DEFAULT)
        optimizer_quantization_parameter = Adam([torch.Tensor([])], lr=LR_QUANTIZATION_PARAM_DEFAULT)
        # TODO: Once implementing Soft Quantizer for GPTQ in Pytorch:
        #  - change default quantization_parameters_learning to True.
        #  - remove explicit rounding_type and quantizer_config (and let it use the default GradientPTQConfig).
        return GradientPTQConfigV2(n_epochs,
                                   optimizer,
                                   optimizer_rest=optimizer_rest,
                                   loss=loss,
                                   log_function=log_function,
                                   train_bias=True,
                                   optimizer_quantization_parameter=optimizer_quantization_parameter,
                                   optimizer_bias=bias_optimizer,
                                   rounding_type=RoundingType.STE,
                                   quantizer_config=GPTQQuantizerConfig(),
                                   quantization_parameters_learning=False,
                                   )


    def pytorch_gradient_post_training_quantization_experimental(model: Module,
                                                                 representative_data_gen: Callable,
                                                                 target_kpi: KPI = None,
                                                                 core_config: CoreConfig = CoreConfig(),
                                                                 gptq_config: GradientPTQConfigV2 = None,
                                                                 gptq_representative_data_gen: Callable = None,
                                                                 target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_PYTORCH_TPC,
                                                                 new_experimental_exporter: bool = False):
        """
        Quantize a trained Pytorch module using post-training quantization.
        By default, the module is quantized using a symmetric constraint quantization thresholds
        (power of two) as defined in the default TargetPlatformCapabilities.
        The module is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        Thresholds are then being calculated using the collected statistics and the module is quantized
        (both coefficients and activations by default).
        If gptq_config is passed, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized modules, and minimizing the
        observed loss.
        Then, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized models, and minimizing the observed
        loss.

        Args:
            model (Module): Pytorch model to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
            core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
            gptq_config (GradientPTQConfigV2): Configuration for using gptq (e.g. optimizer).
            gptq_representative_data_gen (Callable): Dataset used for GPTQ training. If None defaults to representative_data_gen
            target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the PyTorch model according to. `Default PyTorch TPC <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/tpc_models/pytorch_tp_models/pytorch_default.py>`_
            new_experimental_exporter (bool): Whether exporting the quantized model using new exporter or not (in progress. Avoiding it for now is recommended).

        Returns:
            A quantized module and information the user may need to handle the quantized module.

        Examples:

            Import a Pytorch module:

            >>> from torchvision import models
            >>> module = models.mobilenet_v2()

            Create a random dataset generator, for required number of calibration iterations (num_calibration_batches):
            In this example a random dataset of 10 batches each containing 4 images is used.

            >>> import numpy as np
            >>> num_calibration_batches = 10
            >>> def repr_datagen():
            >>>     for _ in range(num_calibration_batches):
            >>>         yield [np.random.random((4, 3, 224, 224))]

            Create MCT core configurations with number of calibration iterations set to 1:

            >>> config = mct.CoreConfig()

            Pass the module, the representative dataset generator and the configuration (optional) to get a quantized module

            >>> quantized_module, quantization_info = mct.pytorch_gradient_post_training_quantization_experimental(module, repr_datagen, core_config=config, gptq_config=gptq_conf)

        """

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use keras_post_training_quantization "
                                    "API, or pass a valid mixed precision configuration.")  # pragma: no cover

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")

        tb_w = _init_tensorboard_writer(DEFAULT_PYTORCH_INFO)

        fw_impl = PytorchImplementation()

        # ---------------------- #
        # Core Runner
        # ---------------------- #
        graph, bit_widths_config = core_runner(in_model=model,
                                               representative_data_gen=representative_data_gen,
                                               core_config=core_config,
                                               fw_info=DEFAULT_PYTORCH_INFO,
                                               fw_impl=fw_impl,
                                               tpc=target_platform_capabilities,
                                               target_kpi=target_kpi,
                                               tb_w=tb_w)

        # ---------------------- #
        # GPTQ Runner
        # ---------------------- #
        graph_gptq = gptq_runner(graph, core_config, gptq_config,
                                 representative_data_gen,
                                 gptq_representative_data_gen if gptq_representative_data_gen else representative_data_gen,
                                 DEFAULT_PYTORCH_INFO, fw_impl, tb_w)
        if core_config.debug_config.analyze_similarity:
            analyzer_model_quantization(representative_data_gen, tb_w, graph_gptq, fw_impl, DEFAULT_PYTORCH_INFO)

        # ---------------------- #
        # Export
        # ---------------------- #
        if new_experimental_exporter:
            Logger.warning('Using new experimental exported models. '
                           'Please do not use unless you are familiar with what you are doing')

            return get_fully_quantized_pytorch_model(graph_gptq)

        return export_model(graph_gptq,
                            DEFAULT_PYTORCH_INFO,
                            fw_impl,
                            tb_w,
                            bit_widths_config)

else:
    # If torch is not installed,
    # we raise an exception when trying to use these functions.
    def get_pytorch_gptq_config(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_gradient_post_training_quantization_experimental. '
                        'Could not find torch package.')  # pragma: no cover


    def pytorch_gradient_post_training_quantization_experimental(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_gradient_post_training_quantization_experimental. '
                        'Could not find the torch package.')  # pragma: no cover



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure import QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.constants import QUANTIZATION_TARGET, \
    QUANTIZATION_METHOD
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.get_all_subclasses import get_all_subclasses


def get_inferable_quantizer_class(quant_target: QuantizationTarget,
                                  quant_method: QuantizationMethod,
                                  quantizer_base_class: type) -> type:
    """
    Searches for an inferable quantizer class that matches the requested QuantizationTarget and QuantizationMethod.
    Exactly one class should be found.

    Args:
        quant_target: QuantizationTarget value (Weights or Activation) which indicates what is the target for
            quantization to use the quantizer for.
        quant_method: A list of QuantizationMethod values to indicate all type of quantization methods that the
            quantizer supports.
        quantizer_base_class: A type of quantizer that the requested quantizer should inherit from.

    Returns: A class of a quantizer that inherits from BaseKerasInferableQuantizer.

    """
    qat_quantizer_classes = get_all_subclasses(quantizer_base_class)
    filtered_quantizers = list(filter(lambda q_class: getattr(q_class, QUANTIZATION_TARGET) == quant_target and
                                                      getattr(q_class, QUANTIZATION_METHOD) is not None and
                                                       quant_method in getattr(q_class, QUANTIZATION_METHOD),
                                      qat_quantizer_classes))

    if len(filtered_quantizers) != 1:
        Logger.error(f"Found {len(filtered_quantizers)} quantizer for target {quant_target.value} "
                     f"that matches the requested quantization method {quant_method.name} "
                     f"but there should be exactly one."
                     f"The possible quantizers that were found are {filtered_quantizers}.")

    return filtered_quantizers[0]



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Dict, Any

from model_compression_toolkit.core.common import BaseNode, Logger
from model_compression_toolkit.core.common.constants import THRESHOLD, SIGNED, RANGE_MIN, RANGE_MAX
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure import QuantizationTarget
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.get_quantizers import \
    get_inferable_quantizer_class
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.pytorch.quantizers import \
    constants as qi_inferable_quantizers_constants, BasePyTorchInferableQuantizer
import numpy as np


def get_weights_inferable_quantizer_kwargs(node: BaseNode) -> Dict[str, Any]:
    # Get the weights quantization configuration for the node
    node_w_qc = node.final_weights_quantization_cfg
    quantization_method = node_w_qc.weights_quantization_method

    # Return the appropriate quantization parameters based on the quantization method
    if quantization_method in [QuantizationMethod.POWER_OF_TWO,
                               QuantizationMethod.SYMMETRIC]:
        return {qi_inferable_quantizers_constants.NUM_BITS: node_w_qc.weights_n_bits,
                qi_inferable_quantizers_constants.THRESHOLD: node_w_qc.weights_quantization_params[THRESHOLD].flatten(),
                qi_inferable_quantizers_constants.PER_CHANNEL: node_w_qc.weights_per_channel_threshold,
                qi_inferable_quantizers_constants.CHANNEL_AXIS: node_w_qc.weights_channels_axis}

    elif quantization_method in [QuantizationMethod.UNIFORM]:
        return {qi_inferable_quantizers_constants.NUM_BITS: node_w_qc.weights_n_bits,
                qi_inferable_quantizers_constants.PER_CHANNEL: node_w_qc.weights_per_channel_threshold,
                qi_inferable_quantizers_constants.MIN_RANGE: node_w_qc.weights_quantization_params[RANGE_MIN].flatten(),
                qi_inferable_quantizers_constants.MAX_RANGE: node_w_qc.weights_quantization_params[RANGE_MAX].flatten(),
                qi_inferable_quantizers_constants.CHANNEL_AXIS: node_w_qc.weights_channels_axis}
    else:
        Logger.critical(f'Not supported quantization method for weights inferable quantizers.')  # pragma: no cover


def get_activation_inferable_quantizer_kwargs(node: BaseNode) -> Dict[str, Any]:
    # Get the activation quantization configuration for the node
    node_qc = node.final_activation_quantization_cfg
    quantization_method = node_qc.activation_quantization_method

    # Return the appropriate quantization parameters based on the quantization method
    if quantization_method in [QuantizationMethod.POWER_OF_TWO,
                               QuantizationMethod.SYMMETRIC]:
        return {qi_inferable_quantizers_constants.NUM_BITS: node_qc.activation_n_bits,
                qi_inferable_quantizers_constants.THRESHOLD: np.asarray([node_qc.activation_quantization_params[THRESHOLD]]),
                qi_inferable_quantizers_constants.SIGNED: node_qc.activation_quantization_params.get(SIGNED)}

    elif quantization_method in [QuantizationMethod.UNIFORM]:
        return {qi_inferable_quantizers_constants.NUM_BITS: node_qc.activation_n_bits,
                qi_inferable_quantizers_constants.MIN_RANGE: np.asarray([node_qc.activation_quantization_params[RANGE_MIN]]),
                qi_inferable_quantizers_constants.MAX_RANGE: np.asarray([node_qc.activation_quantization_params[RANGE_MAX]])}
    else:
        Logger.critical(f'Not supported quantization method for inferable quantizers.')  # pragma: no cover


def get_weights_quantizer_for_node(node: BaseNode) -> BasePyTorchInferableQuantizer:
    """
    Get weights quantizer for a node.

    Args:
        node: Node to create a weight quantizer for.

    Returns:
        Quantizer for the node's weights.

    """
    if node.final_weights_quantization_cfg is None:
        Logger.critical(f'Can not set quantizer for a node with no final weights quantization configuration')  # pragma:
        # no cover
    node_w_qc = node.final_weights_quantization_cfg
    weights_quantization_method = node_w_qc.weights_quantization_method

    quantier_for_node = get_inferable_quantizer_class(QuantizationTarget.Weights,
                                                      weights_quantization_method,
                                                      BasePyTorchInferableQuantizer)
    kwargs = get_weights_inferable_quantizer_kwargs(node)

    return quantier_for_node(**kwargs)


def get_activations_quantizer_for_node(node: BaseNode) -> BasePyTorchInferableQuantizer:
    """
    Get activation quantizer for a node.

    Args:
        node: Node to create an activation quantizer for.

    Returns:
        Quantizer for the node's activations.

    """
    if node.final_activation_quantization_cfg is None:
        Logger.critical(f'Can not set quantizer for a node with no final activation quantization configuration')  #
        # pragma: no cover
    node_act_qc = node.final_activation_quantization_cfg
    activation_quantization_method = node_act_qc.activation_quantization_method

    quantier_for_node = get_inferable_quantizer_class(QuantizationTarget.Activation,
                                                      activation_quantization_method,
                                                      BasePyTorchInferableQuantizer)
    kwargs = get_activation_inferable_quantizer_kwargs(node)

    return quantier_for_node(**kwargs)




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import FOUND_TF
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.get_all_subclasses import get_all_subclasses

if FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit import quantizers_infrastructure as qi
    from model_compression_toolkit.quantizers_infrastructure import BaseKerasTrainableQuantizer
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.base_keras_inferable_quantizer \
        import \
        BaseKerasInferableQuantizer
    keras = tf.keras

    def keras_load_quantized_model(filepath, custom_objects=None, compile=True, options=None):
        """
        This function wraps the keras load model and MCT quantization custom class to it.

        Args:
            filepath: the model file path.
            custom_objects: Additional custom objects
            compile: Boolean, whether to compile the model after loading.
            options: Optional `tf.saved_model.LoadOptions` object that specifies options for loading from SavedModel.

        Returns: A keras Model

        """
        qi_inferable_custom_objects = {subclass.__name__: subclass for subclass in
                                       get_all_subclasses(BaseKerasInferableQuantizer)}
        all_inferable_names = list(qi_inferable_custom_objects.keys())
        if len(set(all_inferable_names)) < len(all_inferable_names):
            Logger.error(f"Found multiple quantizers with the same name that inherit from BaseKerasInferableQuantizer"
                         f"while trying to load a model.")

        qi_trainable_custom_objects = {subclass.__name__: subclass for subclass in
                                       get_all_subclasses(BaseKerasTrainableQuantizer)}
        all_trainable_names = list(qi_trainable_custom_objects.keys())
        if len(set(all_trainable_names)) < len(all_trainable_names):
            Logger.error(f"Found multiple quantizers with the same name that inherit from BaseKerasTrainableQuantizer"
                         f"while trying to load a model.")

        # Merge dictionaries into one dict
        qi_custom_objects = {**qi_inferable_custom_objects, **qi_trainable_custom_objects}

        # Add non-quantizers custom objects
        qi_custom_objects.update({qi.KerasQuantizationWrapper.__name__: qi.KerasQuantizationWrapper})
        if custom_objects is not None:
            qi_custom_objects.update(custom_objects)
        return tf.keras.models.load_model(filepath,
                                          custom_objects=qi_custom_objects, compile=compile,
                                          options=options)
else:
    def keras_load_quantized_model(filepath, custom_objects=None, compile=True, options=None):
        """
        This function wraps the keras load model and MCT quantization custom class to it.

        Args:
            filepath: the model file path.
            custom_objects: Additional custom objects
            compile: Boolean, whether to compile the model after loading.
            options: Optional `tf.saved_model.LoadOptions` object that specifies options for loading from SavedModel.

        Returns: A keras Model

        """
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_load_quantized_model. '
                        'Could not find Tensorflow package.')  # pragma: no cover



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, List


def get_threshold_reshape_shape(tensor_shape: Tuple, quant_axis: int, quant_axis_dim: int) -> List[int]:
    """
    Gets a shape that contains 1 in all axis except the quantization axis, to adjust the threshold tensor for
    per-channel quantization.

    Args:
        tensor_shape: The shape of th

        e tensor to be quantized.
        quant_axis: The axis along which the quantization happens (usually the tensor's channel axis).
        quant_axis_dim: The dimension of the quantization axis.

    Returns: A shape to reshape the threshold tensor according to.

    """
    n_axis = len(tensor_shape)
    quantization_axis = n_axis + quant_axis if quant_axis < 0 else quant_axis

    return [quant_axis_dim if i == quantization_axis else 1 for i in range(n_axis)]



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TORCH
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer, \
    QuantizationTarget

if FOUND_TORCH:
    import torch
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.pytorch.quantizers.base_symmetric_inferable_quantizer import \
        BaseSymmetricInferableQuantizer


    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.SYMMETRIC],
                    quantizer_type=None)
    class ActivationSymmetricInferableQuantizer(BaseSymmetricInferableQuantizer):
        """
        Class for quantizing activations using a symmetric quantizer
        """

        def __init__(self,
                     num_bits: int,
                     threshold: np.ndarray,
                     signed: bool):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing activations
                signed: whether to use signed quantization or not
            """

            super(ActivationSymmetricInferableQuantizer, self).__init__(
                num_bits=num_bits,
                threshold=threshold,
                signed=signed)

            # Activation supports only per-tensor quantization
            assert len(
                self.scales) == 1, f'For activation, quantization per channel is not supported and threshold should ' \
                                   f'be of length 1 but is {len(threshold)}'
            self.scales = self.scales[0]

            self.zero_points = 0

        def __call__(self, inputs: torch.Tensor):
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            return torch.fake_quantize_per_tensor_affine(inputs,
                                                         scale=self.scales,
                                                         zero_point=self.zero_points,
                                                         quant_min=self.min_quantized_domain,
                                                         quant_max=self.max_quantized_domain)

else:
    class ActivationSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using ActivationSymmetricInferableQuantizer. '
                            'Could not find torch package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Set


def get_all_subclasses(cls: type) -> Set[type]:
    """
    This function returns a list of all subclasses of the given class,
    including all subclasses of those subclasses, and so on.
    Recursively get all subclasses of the subclass and add them to the list of all subclasses.

    Args:
        cls: A class object.

    Returns: All classes that inherit from cls.

    """

    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)])



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any, Tuple

import numpy as np

from model_compression_toolkit.core.common.constants import EPS

#########################
#  Helpful functions
#########################


def validate_before_compute_similarity(float_tensor: Any, fxp_tensor: Any):
    """
    Assert different conditions before comparing two tensors (such as dimensionality and type).
    Args:
        float_tensor: First tensor to validate.
        fxp_tensor: Second tensor to validate.

    """
    assert isinstance(float_tensor, np.ndarray)
    assert isinstance(fxp_tensor, np.ndarray)
    assert float_tensor.shape == fxp_tensor.shape


def _similarity_tensor_norm(x: np.ndarray, p: float = 2.0) -> np.ndarray:
    """
    Compute the Lp-norm of a tensor x.
    Args:
        x: Tensor to compute its norm
        p: P to use for the Lp norm computation

    Returns:
        Lp norm per sample in batch of tensor x.
    """

    return (np.abs(x) ** p).sum(axis=-1) ** (1.0/p)


def flatten_tensor(t: np.ndarray, batch: bool) -> np.ndarray:
    """
    Flattening the samples batch to allow similarity analysis computation per sample.

    Args:
        t: A tensor to be flattened.
        batch: Whether the similarity computation is per image or per tensor.

    Returns: A flattened tensor which has the number of samples as is first dimension.

    """

    if batch:
        f_t = t.reshape([t.shape[0], -1])
    else:
        f_t = t.flatten()

    return f_t

#########################
# Similarity functions
#########################


def compute_mse(float_tensor: np.ndarray,
                fxp_tensor: np.ndarray,
                norm: bool = False,
                norm_eps: float = 1e-8,
                batch: bool = False) -> float:
    """
    Compute the mean square error between two numpy arrays.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        norm: whether to normalize the error function result.
        norm_eps: epsilon value for error normalization stability.
        batch: Whether to run batch similarity analysis or not.

    Returns:
        The MSE distance between the two tensors.
    """
    validate_before_compute_similarity(float_tensor, fxp_tensor)

    float_flat = flatten_tensor(float_tensor, batch)
    fxp_flat = flatten_tensor(fxp_tensor, batch)

    error = ((float_flat - fxp_flat) ** 2).mean(axis=-1)
    if norm:
        error /= ((float_flat ** 2).mean(axis=-1) + norm_eps)

    return error


def compute_mae(float_tensor: np.ndarray,
                fxp_tensor: np.ndarray,
                norm: bool = False,
                norm_eps: float = 1e-8,
                batch: bool = False) -> float:
    """
    Compute the mean average error function between two numpy arrays.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        norm: whether to normalize the error function result.
        norm_eps: epsilon value for error normalization stability.
        batch: Whether to run batch similarity analysis or not.

    Returns:
        The mean average distance between the two tensors.
    """

    validate_before_compute_similarity(float_tensor, fxp_tensor)

    float_flat = flatten_tensor(float_tensor, batch)
    fxp_flat = flatten_tensor(fxp_tensor, batch)

    error = np.abs(float_flat - fxp_flat).mean(axis=-1)
    if norm:
        error /= (np.abs(float_flat).mean(axis=-1) + norm_eps)
    return error


def compute_cs(float_tensor: np.ndarray, fxp_tensor: np.ndarray, eps: float = 1e-8, batch: bool = False) -> float:
    """
    Compute the similarity between two tensor using cosine similarity.
    The returned values is between 0 to 1: the smaller returned value,
    the greater similarity there is between the two tensors.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        eps: Small value to avoid zero division.
        batch: Whether to run batch similarity analysis or not.

    Returns:
        The cosine similarity between two tensors.
    """

    validate_before_compute_similarity(float_tensor, fxp_tensor)
    if np.all(fxp_tensor == 0) and np.all(float_tensor == 0):
        return 1.0

    float_flat = flatten_tensor(float_tensor, batch)
    fxp_flat = flatten_tensor(fxp_tensor, batch)

    float_norm = _similarity_tensor_norm(float_flat)
    fxp_norm = _similarity_tensor_norm(fxp_flat)

    # -1 <= cs <= 1
    axis = None if not batch else 1
    cs = np.sum(float_flat * fxp_flat, axis=axis) / ((float_norm * fxp_norm) + eps)

    # Return a non-negative float (smaller value -> more similarity)
    return (1.0 - cs) / 2.0


def compute_lp_norm(float_tensor: np.ndarray,
                    fxp_tensor: np.ndarray,
                    p: int,
                    norm: bool = False,
                    norm_eps: float = 1e-8,
                    batch: bool = False) -> float:
    """
    Compute the error function between two numpy arrays.
    The error is computed based on Lp-norm distance of the tensors.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        p: p-norm to use for the Lp-norm distance.
        norm: whether to normalize the error function result.
        norm_eps: epsilon value for error normalization stability.
        batch: Whether to run batch similarity analysis or not.

    Returns:
        The Lp-norm distance between the two tensors.
    """
    validate_before_compute_similarity(float_tensor, fxp_tensor)

    float_flat = flatten_tensor(float_tensor, batch)
    fxp_flat = flatten_tensor(fxp_tensor, batch)

    error = (np.abs(float_flat - fxp_flat) ** p).mean(axis=-1)
    if norm:
        error /= ((np.abs(float_flat) ** p).mean(axis=-1) + norm_eps)
    return error


def compute_kl_divergence(float_tensor: np.ndarray, fxp_tensor: np.ndarray, batch: bool = False) -> float:
    """
    Compute the similarity between two tensor using KL-divergence.
    The returned values is between 0 to 1: the smaller returned value,
    the greater similarity there is between the two tensors.

    Args:
        float_tensor: First tensor to compare.
        fxp_tensor: Second tensor to compare.
        batch: Whether to run batch similarity analysis or not.

    Returns:
        The KL-divergence between two tensors.
    """

    validate_before_compute_similarity(float_tensor, fxp_tensor)

    float_flat = flatten_tensor(float_tensor, batch)
    fxp_flat = flatten_tensor(fxp_tensor, batch)

    non_zero_fxp_tensor = fxp_flat.copy()
    non_zero_fxp_tensor[non_zero_fxp_tensor == 0] = EPS

    return np.sum(np.where(float_flat != 0, float_flat * np.log(float_flat / non_zero_fxp_tensor), 0), axis=-1)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
from typing import Callable

import numpy as np

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.base_node import BaseNode
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common.target_platform import QuantizationMethod


class BatchNormalizationReconstruction(common.BaseSubstitution):
    """
    Reconstruct BatchNormalization after linear layers.
    """
    def __init__(self,
                 source_node: NodeOperationMatcher,
                 create_bn_node: Callable,
                 gamma_str: str,
                 beta_str: str,
                 moving_mean_str: str,
                 moving_variance_str: str,
                 epsilon_val: float):
        """
        Matches: Source (linear) nodes.

        Args:
            source_node: Node matcher for linear type nodes.
            create_bn_node: Function for creating BN nodes.
            gamma_str: The framework specific attribute name of the batch norm layer's gamma parameter.
            beta_str: The framework specific attribute name of the batch norm layer's beta parameter.
            moving_mean_str: The framework specific attribute name of the batch norm layer's moving mean parameter.
            moving_variance_str: The framework specific attribute name of the batch norm layer's moving
            variance parameter.
            epsilon_val: The framework specific attribute value of the batch norm layer's epsilon parameter.
        """
        super().__init__(matcher_instance=source_node)
        self.create_bn_node = create_bn_node
        self.gamma_str = gamma_str
        self.beta_str = beta_str
        self.moving_mean_str = moving_mean_str
        self.moving_variance_str = moving_variance_str
        self.epsilon_val = epsilon_val

    def substitute(self,
                   graph: Graph,
                   source_node: BaseNode) -> Graph:
        """
        Reconstruct BatchNormalization after linear layers.

        Args:
            graph: Graph we apply the substitution on.
            source_node: linear type node.

        Returns:
            Graph after applying the substitution.
        """

        num_nodes_before_substitution = len(graph.nodes)
        num_edges_before_substitution = len(graph.edges)

        # If the linear operator is part of a reused group (it is the "base" node, or a reused node),
        # we should skip the substitution.
        if source_node.reuse or source_node.reuse_group is not None:
            for qc in source_node.candidates_quantization_cfg:
                qc.weights_quantization_cfg.weights_second_moment_correction = False
            return graph

        # We apply only on nodes with folded BatchNormalization.
        if source_node.prior_info.std_output is None or source_node.prior_info.mean_output is None:
            for qc in source_node.candidates_quantization_cfg:
                qc.weights_quantization_cfg.weights_second_moment_correction = False
            return graph

        # This feature disabled for models with weights quantization method of Power of 2
        for qc in source_node.candidates_quantization_cfg:
            if qc.weights_quantization_cfg.weights_quantization_method == QuantizationMethod.POWER_OF_TWO:
                Logger.warning("Second moment statistics correction feature disabled for models with weights "
                               "quantization method of Power of 2")
                for qc_inner in source_node.candidates_quantization_cfg:
                    qc_inner.weights_quantization_cfg.weights_second_moment_correction = False
                return graph

        eps = self.epsilon_val

        original_gamma = source_node.prior_info.std_output
        beta = source_node.prior_info.mean_output
        moving_mean = beta
        moving_var = np.power(original_gamma, 2)
        gamma = np.sqrt(moving_var + eps)

        bn_node_weights = {self.gamma_str: gamma,
                           self.beta_str: beta,
                           self.moving_mean_str: moving_mean,
                           self.moving_variance_str: moving_var}

        bn_node = self.create_bn_node(source_node, bn_node_weights)

        bn_node.prior_info = copy.deepcopy(source_node.prior_info)

        bn_node.candidates_quantization_cfg = copy.deepcopy(source_node.candidates_quantization_cfg)

        for qc in bn_node.candidates_quantization_cfg:
            qc.weights_quantization_cfg.enable_weights_quantization = False
            qc.activation_quantization_cfg.enable_activation_quantization = False

        graph.reconnect_out_edges(current_node=source_node, new_node=bn_node)
        graph.replace_output_node(current_node=source_node, new_node=bn_node)
        graph.add_node_with_in_edges(bn_node, [source_node])

        assert len(graph.nodes) - num_nodes_before_substitution == 1
        assert len(graph.edges) - num_edges_before_substitution == 1

        return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from typing import Any, Callable

from tensorflow.keras.layers import BatchNormalization
from tqdm import tqdm

import model_compression_toolkit.core.keras.constants as keras_constants
from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core import common


def keras_apply_second_moment_correction(quantized_model: Any,
                                         core_config: CoreConfig,
                                         representative_data_gen: Callable,
                                         graph: common.Graph):
    """
    Apply second moment statistics correction to graph.

    Args:
        quantized_model: Framework's model to apply second moment correction on.
        core_config: QuantizationConfig of how the model should be quantized.
        representative_data_gen: Dataset to use for retrieving images for the models inputs.
        graph: Graph to update the parameters after the second moment correction.

    Returns:
        A function that applies second moment correction.
    """

    # Move every BN to train mode
    for layer in quantized_model.layers:
        if len(graph.find_node_by_name(layer.name)) > 0:
            node = graph.find_node_by_name(layer.name)[0]
            if isinstance(layer, BatchNormalization) and node.final_weights_quantization_cfg\
                    .weights_second_moment_correction:
                layer.trainable = True

    for data in tqdm(representative_data_gen()):
        quantized_model(data, training=True)

    # Move every BN to eval mode and update the corresponding BN node params in the graph
    for layer in quantized_model.layers:
        if len(graph.find_node_by_name(layer.name)) > 0:
            node = graph.find_node_by_name(layer.name)[0]
            if isinstance(layer, BatchNormalization) and node.final_weights_quantization_cfg\
                    .weights_second_moment_correction:
                layer.trainable = False
                bn_node_weights = {keras_constants.GAMMA: layer.gamma,
                                   keras_constants.BETA: layer.beta,
                                   keras_constants.MOVING_MEAN: layer.moving_mean,
                                   keras_constants.MOVING_VARIANCE: layer.moving_variance}
                node.weights = copy.deepcopy(bn_node_weights)
    return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, Callable
import torch

from model_compression_toolkit.core.common.constants import THRESHOLD, SIGNED, RANGE_MIN, RANGE_MAX
from model_compression_toolkit.core.common.quantization.quantizers.uniform_quantizers import threshold_is_power_of_two
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import fix_range_to_include_zero


def get_symmetric_quantization_range_and_scale(activation_is_signed: bool,
                                               activation_n_bits: int,
                                               activation_threshold: float):
    """
    Calculates lower and upper bounds on the quantization range, along with quantization scale,
    for symmetric quantization (used for the symmetric and power-of-two quantizers),
    according to whether the quantization is signed or unsigned.

    Args:
        activation_is_signed: Whether the quantization is signed or not.
        activation_n_bits: Number of bits to use for quantization.
        activation_threshold: The quantization threshold.

    Returns: range lower bound, range upper bound and quantization scale.

    """
    if activation_is_signed:
        min_value = -2 ** (activation_n_bits - 1)
        max_value = 2 ** (activation_n_bits - 1) - 1
        scale = activation_threshold / 2 ** (activation_n_bits - 1)
    else:
        min_value = 0
        max_value = (2 ** activation_n_bits) - 1
        scale = activation_threshold / 2 ** activation_n_bits

    return min_value, max_value, scale


def power_of_two_quantization(activation_n_bits: int,
                              quantization_params: dict) -> Callable:
    """
    Use a NodeQuantizationConfig to compute a quantizer min/max values, and use it to
    build and return a fake-quantization node, quantized with a power-of-two threshold.
    Args:
        activation_n_bits: Number of bits to use for quantization.
        quantization_params: Dictionary of specific parameters for this quantization function.
    Returns:
        A fake quantization node.
    """
    activation_threshold = quantization_params.get(THRESHOLD)
    activation_is_signed = quantization_params.get(SIGNED)

    if activation_threshold is None or activation_is_signed is None:
        return None # pragma: no cover
    if not threshold_is_power_of_two(activation_threshold, per_channel=False):
        return None # pragma: no cover

    min_value, max_value, scale = get_symmetric_quantization_range_and_scale(activation_is_signed,
                                                                             activation_n_bits,
                                                                             activation_threshold)

    return lambda x: q(x, min_value, max_value, scale)


def symmetric_quantization(activation_n_bits: int,
                           quantization_params: dict) -> Callable:
    """
    Use a NodeQuantizationConfig to compute a quantizer min/max values, and use it to
    build and return a symmetric fake-quantization node.

    Args:
        activation_n_bits: Number of bits to use for quantization.
        quantization_params: Dictionary of specific parameters for this quantization function.

    Returns:
        A fake quantization node.
    """
    activation_threshold = quantization_params.get(THRESHOLD)
    activation_is_signed = quantization_params.get(SIGNED)

    if activation_threshold is None or activation_is_signed is None:
        return None # pragma: no cover

    min_value, max_value, scale = get_symmetric_quantization_range_and_scale(activation_is_signed,
                                                                             activation_n_bits,
                                                                             activation_threshold)

    return lambda x: q(x, min_value, max_value, scale)


def uniform_quantization(activation_n_bits: int,
                         quantization_params: dict) -> Callable:
    """
    Use a NodeQuantizationConfig to compute a quantizer min/max values, and use it to
    build and return a uniform fake-quantization node.

    Args:
        activation_n_bits: Number of bits to use for quantization.
        quantization_params: Dictionary of specific parameters for this quantization function.

    Returns:
        A fake quantization node.
    """
    a, b = quantization_params.get(RANGE_MIN), quantization_params.get(RANGE_MAX)

    if a is None or b is None:
        return None # pragma: no cover

    # fixing quantization range to include 0
    a = 0 if a > 0 else a
    b = 0 if b < 0 else b
    a, b = fix_range_to_include_zero(a, b, activation_n_bits)

    min_value = 0
    max_value = 2 ** activation_n_bits - 1
    scale = (b - a) / ((2 ** activation_n_bits) - 1)
    zero_point = -round(a / scale)  # zp has to be positive, and a <=0, so we multiply by -1

    return lambda x: q(x, min_value, max_value, scale, zero_point)


def q(x: torch.Tensor,
      min_value: int,
      max_value: int,
      scale: float,
      zero_point: int = 0) -> torch.Tensor:
    """
    Fake-quantize the input tensor x, using a pytorch fake-quantization node.
    Args:
        x: input tensor to quantize.
        min_value: lower bound of the quantized domain.
        max_value: upper bound of the quantized domain.
        scale: quantization scale.
        zero_point: quantization zero_point
    Returns:
        The fake-quantized input tensor.
    """

    return torch.fake_quantize_per_tensor_affine(x,
                                                 scale=scale,
                                                 zero_point=zero_point,
                                                 quant_min=min_value,
                                                 quant_max=max_value)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
import torch
from typing import Any, List

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.core.pytorch.back2framework.quantization_wrapper.wrapper_quantize_config import \
    WrapperQuantizeConfig
from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
from model_compression_toolkit.core.pytorch.utils import set_model, to_torch_tensor


class QuantizedLayerWrapper(torch.nn.Module):
    """
    Class that wraps a Pytorch layer (nn.Module).
    The wrapper creates an inside PyTorch module and wraps it using a WrapperQuantizationConfig
    which defines how the layer should be quantized.
    """
    def __init__(self,
                 n: BaseNode,
                 quantize_config:WrapperQuantizeConfig):
        """
        Construct a Pytorch model that constitutes as a wrapper for a Pytorch layer, built from a given graph node.

        Args:
            n: Node to build its PyTorch layer.
            quantize_config: WrapperQuantizeConfig to define how the layer is quantized.

        """
        super(QuantizedLayerWrapper, self).__init__()

        # Is layer a function or a layer
        self.is_function = isinstance(n, FunctionalNode)

        # Build a layer and load its weights
        self._build_layer(n)
        self.quantization_config = quantize_config

        # Setting layers' weights
        if self.quantization_config.is_weight_quantized and not self.is_function:
            self._quantize_weights(n)

        if not self.is_function:
            set_model(self.layer)

    def forward(self, x: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Args:
            x: Input tensors to layer.

        Returns:
            Output tensor of the wrapped layer.
        """

        outputs = self.layer(x, *args, **kwargs)

        if self.quantization_config.is_activation_quantized:
            if isinstance(outputs, list):
                outputs = torch.cat(outputs, dim=0)
            outputs = self.quantization_config.get_activation_quantizers()[0](outputs)

        return outputs

    def _build_layer(self, n: BaseNode):
        """
        Build and set the inner layer which QuantizedLayerWrapper wraps.

        Args:
            n: Node to build its PyTorch layer.

        Returns:
            None.
        """
        if self.is_function:
            self.layer = n.type
        else:
            framework_attr = copy.copy(n.framework_attr)
            self.layer = n.type(**framework_attr)
            self.layer.load_state_dict({k: torch.Tensor(v) for k, v in n.weights.items()}, strict=False)

    def _quantize_weights(self, n:BaseNode):
        """
        Quantize node's weights and load them as the layer's weights.

        Args:
            n: Node to quantize its weights.

        Returns:
            None.
        """

        self.weight_attrs = DEFAULT_PYTORCH_INFO.get_kernel_op_attributes(n.type)

        # float_weights is a list of weights for each attribute that we want to quantize.
        float_weights = [n.get_weights_by_keys(attr) for attr in self.weight_attrs]
        assert len(self.weight_attrs) == len(float_weights)

        # Quantize weights and load them to layer.
        self.quantized_weights = self._get_quantized_weights(float_weights)
        if not self.is_function:
            self._load_quantized_weights()

    def _load_quantized_weights(self):
        """
        Load quantized weights of a layer.
        """

        # loading the weights (if exists) from the graph node (weights of the trained model)
        assert len(self.weight_attrs) == len(self.quantized_weights)
        loaded_weights = {k: torch.as_tensor(v) for k, v in self.layer.state_dict().items()}

        with torch.no_grad():
            for attr_idx, attr in enumerate(self.weight_attrs):
                # need to prepare the weights' tensor - extract it from the maintained quantized_weights list
                # and move it to the relevant device as the wrapped layer's weights.
                weights_tensor = self.quantized_weights[attr_idx]
                active_weights = torch.nn.Parameter(to_torch_tensor(weights_tensor))
                loaded_weights[self.weight_attrs[attr_idx]] = active_weights
            self.layer.load_state_dict(loaded_weights, strict=False)

    def _get_quantized_weights(self, float_weights: List[torch.Tensor]) -> List[torch.Tensor]:
        """
        Get quantized weights' tensors.

        Returns:
            List of quantized weights (for each layer's attribute to be quantized).
        """

        quantized_weights = []
        quantizers = self.quantization_config.get_weight_quantizers()
        assert len(quantizers)==len(float_weights)
        for float_weight, quantizer in zip(float_weights, quantizers):
            # for each attribute
            quantized_weights.append(quantizer(float_weight=float_weight))
        return quantized_weights





from typing import Dict, Callable

import torch
import numpy as np

from model_compression_toolkit.core.common.constants import SIGNED, CLUSTER_CENTERS, THRESHOLD, MULTIPLIER_N_BITS, EPS
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor


def activation_lut_kmean_quantizer(activation_n_bits: int,
                                   quantization_params: Dict[str, np.ndarray]) -> Callable:
    """
    Builds a LUT quantizer for layer's activation using the provided params (threshold and clusters).
    It initiates a fake custom LUT layer that provides the quantizer function.

    Args:
        activation_n_bits: Number of bits to use for quantization (not used in this function).
        quantization_params: Dictionary of specific parameters for this quantization function.

    Returns:
        A fake LUT quantization node.
    """

    lut_fake_quant = PytorchLUTFakeQuant(quantization_params=quantization_params)
    return lambda x: lut_fake_quant(x)


class PytorchLUTFakeQuant(torch.nn.Module):
    """
    A custom PyTorch layer for quantizing activation tensor with non-uniform quantization (using lookup table clustering).
    """

    def __init__(self,
                 quantization_params: Dict[str, np.ndarray]):
        """
        Construct a Pytorch module that quantizes an activation tensor.

        Args:
            quantization_params: Dictionary of specific parameters for this quantization function.
        """

        super(PytorchLUTFakeQuant, self).__init__()

        self.quantization_params = quantization_params
        self.activation_is_signed = self.quantization_params.get(SIGNED)
        self.cluster_centers = to_torch_tensor(self.quantization_params.get(CLUSTER_CENTERS))
        self.threshold = self.quantization_params.get(THRESHOLD)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Quantize the output of an activation,

        Args:
            x: input tensor (which is the activation output of the layer that we want to quantize)

        Returns:
            Quantized torch Tensor.
        """
        if self.activation_is_signed is None or self.cluster_centers is None or self.threshold is None:
            return None   # pragma: no cover

        _quant_output = self.lut_kmeans_quantizer(x)
        return _quant_output

    def lut_kmeans_quantizer(self, tensor_data: torch.Tensor) -> torch.Tensor:
        """
        Quantize a tensor using a non-uniform quantization based on the pre-defined kmeans clusters.
        1. Scales tensor_data with the threshold into 8-bit quantization range.
        2. Assigns cluster centers to each value.
        3. Scales back by multiplying the result by threshold and dividing with the quantization range max value.
        The result is the quantized tensor.

        Args:
            tensor_data: Input activation tensor.

        Returns: Quantized tensor.
        """

        tensor = self.int_quantization_with_threshold(tensor_data, MULTIPLIER_N_BITS)
        tensor = tensor.unsqueeze(-1)

        expanded_cluster_centers = self.cluster_centers.reshape([*[1 for _ in range(len(tensor.shape) - 1)], -1])
        cluster_assignments = torch.argmin(torch.abs(tensor - expanded_cluster_centers), dim=-1)
        centers = self.cluster_centers.flatten()[cluster_assignments]

        quant_tensor = (centers / (2 ** (MULTIPLIER_N_BITS - int(self.activation_is_signed)))) * self.threshold

        return quant_tensor

    def int_quantization_with_threshold(self,
                                        data: torch.Tensor,
                                        n_bits: int,
                                        eps: float = EPS) -> torch.Tensor:
        """
        Divides data by threshold and quantize it to integers in the quantization range (depends on signed value).

        Args:
            data: tensor data.
            n_bits: number of bits that determines the quantization range.
            eps: Small value for numerical stability in division.

        Returns:
            Uniform Quantized tensor.

        """

        if self.activation_is_signed:
            clip_max = 2 ** (n_bits - 1) - 1
            clip_min = -2 ** (n_bits - 1)
        else:
            clip_max = 2 ** n_bits - 1
            clip_min = 0

        return torch.clip((data / (self.threshold + eps)) * (2 ** (n_bits - int(self.activation_is_signed))),
                          min=clip_min, max=clip_max)



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import tensorflow as tf


def int_quantization_with_threshold(data: tf.Tensor,
                                    n_bits: int,
                                    signed: bool,
                                    threshold: np.ndarray,
                                    eps: float) -> tf.Tensor:
    """
    Divides data by threshold and quantize it to integers in the quantization range (depends on signed value).

    Args:
        data: tensor data.
        n_bits: number of bits that determines the quantization range.
        signed: Whether the quantization is signed or not.
        threshold: threshold for quantization.
        eps: Small value for numerical stability in division.

    Returns:
        Uniform Quantized tensor.

    """

    if signed:
        clip_max = 2 ** (n_bits - 1) - 1
        clip_min = -2 ** (n_bits - 1)
    else:
        clip_max = 2 ** n_bits - 1
        clip_min = 0

    return tf.clip_by_value((data / (threshold + eps)) * (2 ** (n_bits - int(signed))),
                            clip_value_max=clip_max, clip_value_min=clip_min)



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TF

from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer, QuantizationTarget

if FOUND_TF:
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers.weights_inferable_quantizers.weights_uniform_inferable_quantizer \
        import WeightsUniformInferableQuantizer

    @mark_quantizer(quantization_target=QuantizationTarget.Weights,
                    quantization_method=[QuantizationMethod.SYMMETRIC],
                    quantizer_type=None)
    class WeightsSymmetricInferableQuantizer(WeightsUniformInferableQuantizer):
        """
        Class for quantizing weights using a symmetric quantizer
        """
        def __init__(self,
                     num_bits: int,
                     threshold: List[float],
                     per_channel: bool,
                     channel_axis: int = None,
                     input_rank: int = None
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                threshold: threshold for quantizing weights
                per_channel: whether to use per-channel quantization
                channel_axis: axis along which to apply per-channel quantization
                input_rank: number of dimensions of input tensor the quantizer quantizes
            """
            assert isinstance(threshold, list), f'Expected threshold to be of type list but is {type(threshold)}'
            assert all([isinstance(x, (float, np.float32, np.float64)) for x in
                        threshold]), f'Expected threshold list to contain float or np.float values but found ' \
                                     f'{[type(x) for x in threshold]}'

            self.threshold = np.asarray(threshold)

            _min_range = -self.threshold
            _max_range = self.threshold - self.threshold / (2 ** (num_bits - 1))

            super(WeightsSymmetricInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                     min_range=list(_min_range),
                                                                     max_range=list(_max_range),
                                                                     per_channel=per_channel,
                                                                     channel_axis=channel_axis,
                                                                     input_rank=input_rank)

        def get_config(self):
            """
            Return a dictionary with the configuration of the quantizer.

            Returns:
                Dictionary with the following keys: 'num_bits', 'threshold', 'per_channel', 'channel_axis'
            """
            return {'num_bits': self.num_bits,
                    'threshold': self.threshold,
                    'per_channel': self.per_channel,
                    'channel_axis': self.channel_axis,
                    'input_rank': self.input_rank}

else:
    class WeightsSymmetricInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing tensorflow and tensorflow_model_optimization is mandatory '
                            'when using WeightsPOTInferableQuantizer. '
                            'Could not find Tensorflow package.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any

from keras.engine.input_layer import InputLayer

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.quantizers_infrastructure import KerasQuantizationWrapper
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import BaseInferableQuantizer



def is_keras_layer_exportable(layer: Any) -> bool:
    """
    Check whether a Keras layer is a valid exportable layer or not.

    Args:
        layer: Keras layer to check if considered to be valid for exporting.

    Returns:
        Check whether a Keras layer is a valid exportable layer or not.
    """
    # Keras Input layers are not wrapped
    if isinstance(layer, InputLayer):
        return True

    valid_layer = isinstance(layer, KerasQuantizationWrapper)
    if not valid_layer:
        Logger.error(
            f'Exportable layer must be wrapped using KerasQuantizationWrapper, but layer {layer.name} is of type '
            f'{type(layer)}') # pragma: no cover

    valid_weights_quantizers = isinstance(layer.weights_quantizers, dict)
    if not valid_weights_quantizers:
        Logger.error(
            f'KerasQuantizationWrapper must have a weights_quantizers but has a '
            f'{type(layer.weights_quantizers)} object') # pragma: no cover

    for _, weights_quantizer in layer.weights_quantizers.items():
        if not isinstance(weights_quantizer, BaseInferableQuantizer):
            Logger.error(
                f'weights_quantizer must be a BaseInferableQuantizer object but has a '
                f'{type(weights_quantizer)} object')  # pragma: no cover

    valid_activation_quantizers = isinstance(layer.activation_quantizers, list)
    if not valid_activation_quantizers:
        Logger.error(
            f'KerasQuantizationWrapper must have a activation_quantizers list but has a '
            f'{type(layer.activation_quantizers)} object') # pragma: no cover

    for activation_quantizers in layer.activation_quantizers:
        if not isinstance(activation_quantizers, BaseInferableQuantizer):
            Logger.error(
                f'activation_quantizers must be a BaseInferableQuantizer object but has a '
                f'{type(activation_quantizers)} object')  # pragma: no cover

    quantizers = layer.activation_quantizers + list(layer.weights_quantizers.values())
    is_valid_quantizers = all([isinstance(x, BaseInferableQuantizer) for x in quantizers])
    if not is_valid_quantizers:
        Logger.error(f'Found a quantizer that is not of type BaseInferableQuantizer') # pragma: no cover

    return True



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Callable

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import FOUND_TORCH


class PyTorchExportMode(Enum):
    FAKELY_QUANT_TORCHSCRIPT = 0
    FAKELY_QUANT_ONNX = 1


if FOUND_TORCH:
    import torch.nn
    from model_compression_toolkit.exporter.model_exporter.pytorch.fakely_quant_onnx_pytorch_exporter import FakelyQuantONNXPyTorchExporter
    from model_compression_toolkit.exporter.model_exporter.pytorch.fakely_quant_torchscript_pytorch_exporter import FakelyQuantTorchScriptPyTorchExporter
    from model_compression_toolkit.exporter.model_wrapper.pytorch.validate_layer import is_pytorch_layer_exportable

    def pytorch_export_model(model: torch.nn.Module,
                             save_model_path: str,
                             repr_dataset: Callable,
                             is_layer_exportable_fn: Callable = is_pytorch_layer_exportable,
                             mode: PyTorchExportMode = PyTorchExportMode.FAKELY_QUANT_TORCHSCRIPT) -> None:
        """
        Export a PyTorch quantized model to a torchscript or onnx model.
        The model will be saved to the path in save_model_path.
        Mode can be used for different exported files. Currently, pytorch_export_model
        supports PyTorchExportMode.FAKELY_QUANT_TORCHSCRIPT (where the exported model
        is in a TorchScript format and its weights and activations are float fakely-quantized values),
        and PyTorchExportMode.FakelyQuantONNX (where the exported model
        is in an ONNX format and its weights and activations are float fakely-quantized values)

        Args:
            model: Model to export.
            is_layer_exportable_fn: Callable to check whether a layer can be exported or not.
            mode: Mode to export the model according to.
            save_model_path: Path to save the model.
            repr_dataset: Representative dataset for tracing the pytorch model (mandatory for exporting it).

        """

        if mode == PyTorchExportMode.FAKELY_QUANT_TORCHSCRIPT:
            exporter = FakelyQuantTorchScriptPyTorchExporter(model,
                                                             is_layer_exportable_fn,
                                                             save_model_path,
                                                             repr_dataset)

        elif mode == PyTorchExportMode.FAKELY_QUANT_ONNX:
            exporter = FakelyQuantONNXPyTorchExporter(model,
                                                      is_layer_exportable_fn,
                                                      save_model_path,
                                                      repr_dataset)

        else:
            Logger.critical(
                f'Unsupported mode was used {mode.name} to export PyTorch model. '
                f'Please see API for supported modes.')  # pragma: no cover

        exporter.export()

else:
    def pytorch_export_model(*args, **kwargs):
        Logger.error('Installing torch is mandatory '
                     'when using pytorch_export_model. '
                     'Could not find PyTorch packages.')  # pragma: no cover



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

from model_compression_toolkit.core.common.constants import FOUND_TORCH
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.common.base_inferable_quantizer import mark_quantizer, \
    QuantizationTarget

if FOUND_TORCH:
    import torch
    from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.pytorch.quantizers.base_uniform_inferable_quantizer import \
        BaseUniformInferableQuantizer


    @mark_quantizer(quantization_target=QuantizationTarget.Activation,
                    quantization_method=[QuantizationMethod.UNIFORM],
                    quantizer_type=None)
    class ActivationUniformInferableQuantizer(BaseUniformInferableQuantizer):
        """
        Class for quantizing activations using an uniform quantizer
        """

        def __init__(self,
                     num_bits: int,
                     min_range: np.ndarray,
                     max_range: np.ndarray,
                     ):
            """
            Initialize the quantizer with the specified parameters.

            Args:
                num_bits: number of bits to use for quantization
                min_range: min range for quantizing activations
                max_range: max range for quantizing activations
            """
            super(ActivationUniformInferableQuantizer, self).__init__(num_bits=num_bits,
                                                                      min_range=min_range,
                                                                      max_range=max_range)

            assert isinstance(min_range,
                              np.ndarray), f'min_range is expected to be numpy array, but is of type {type(min_range)}'
            assert isinstance(max_range,
                              np.ndarray), f'max_range is expected to be numpy array, but is of type {type(max_range)}'
            assert min_range.ndim == 1, f'min_range is expected to be flatten, but of shape {min_range.shape}'
            assert max_range.ndim == 1, f'max_range is expected to be flatten, but of shape {min_range.shape}'

            assert len(
                min_range) == 1, f'For activation, quantization per channel is not supported and min_range should be ' \
                                 f'of length 1 but is {len(min_range)}'
            assert len(
                max_range) == 1, f'For activation, quantization per channel is not supported and max_range should be ' \
                                 f'of length 1 but is {len(max_range)}'

            # Activation is per-tensor thus we expect only a single min/max values
            min_range = min_range[0]
            max_range = max_range[0]

            # fixing quantization range to include 0
            a = 0 if min_range > 0 else min_range
            b = 0 if max_range < 0 else max_range

            self.scale = float((b - a) / ((2 ** num_bits) - 1))
            self.zero_point = int(-np.round(a / self.scale))  # zp has to be positive, and a <=0, so we multiply by -1

        def __call__(self, inputs: torch.Tensor):
            """
            Quantize the given inputs using the quantizer parameters.

            Args:
                inputs: input tensor to quantize

            Returns:
                quantized tensor.
            """
            return torch.fake_quantize_per_tensor_affine(inputs,
                                                         scale=self.scale,
                                                         zero_point=self.zero_point,
                                                         quant_min=self.min_quantized_domain,
                                                         quant_max=self.max_quantized_domain)


else:
    class ActivationUniformInferableQuantizer:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            raise Exception('Installing torch is mandatory '
                            'when using ActivationUniformInferableQuantizer. '
                            'Could not find torch package.')



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from typing import Callable

import keras.models
import numpy as np
import tensorflow as tf
from keras import Sequential
from keras.layers import Dense, Conv2D, Reshape
from keras.models import clone_model

from model_compression_toolkit import quantizers_infrastructure as qi
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.exporter.model_exporter.keras.fakely_quant_keras_exporter import FakelyQuantKerasExporter
from model_compression_toolkit.quantizers_infrastructure.inferable_infrastructure.keras.quantizers import \
    constants as keras_inferable_constants

BIAS_INITIALIZER = 'bias_initializer'
BIAS_REGULARIZER = 'bias_regularizer'
BIAS_CONSTRAINT = 'bias_constraint'
ACTIVITY_REGULARIZER = 'activity_regularizer'
KERNEL_INITIALIZER = 'kernel_initializer'
KERNEL_REGULARIZER = 'kernel_regularizer'
KERNEL_CONSTRAINT = 'kernel_constraint'
KERNEL_SIZE = 'kernel_size'
PADDING = 'padding'
STRIDES = 'strides'
LAYER_NAME = 'name'
TRAINABLE = 'trainable'
ACTIVATION = 'activation'
USE_BIAS = 'use_bias'
FILTERS = 'filters'
UNITS = 'units'
PAD_VALID = 'valid'
KERNEL = 'kernel'

CONV_KERNEL_CHANNEL_AXIS = 3
CONV_INPUT_CHANNELS_DIM = 4

class INT8TFLiteExporter(FakelyQuantKerasExporter):
    """
    Exporter for INT8 TFLite models.
    The exporter expects to receive an exportable model (where each layer's full quantization parameters
    can be retrieved), and convert it into a quantized model where weights and activations are represented
    as integer data type.
    """

    def __init__(self,
                 model: keras.models.Model,
                 is_layer_exportable_fn: Callable,
                 save_model_path: str):
        """

        Args:
            model: Model to export.
            is_layer_exportable_fn: Callable to check whether a layer can be exported or not.
            save_model_path: Path to save the exported model.
        """
        super().__init__(model,
                         is_layer_exportable_fn,
                         save_model_path)

        self.exported_model = None

    def _get_pointwise_layer_to_replace_dense(self, wrapped_layer: qi.KerasQuantizationWrapper) -> keras.layers.Layer:
        # First we create a pointwise configuration based on the Dense layer's configuration
        dense_cfg = wrapped_layer.layer.get_config()

        # List of pw attributes that should be taken from the dense layer as they are.
        pw_attr_list = [LAYER_NAME, ACTIVATION, USE_BIAS, BIAS_CONSTRAINT,
                        BIAS_INITIALIZER, BIAS_REGULARIZER, TRAINABLE, ACTIVITY_REGULARIZER,
                        KERNEL_INITIALIZER, KERNEL_REGULARIZER, KERNEL_CONSTRAINT]

        pw_cfg = {attr: dense_cfg[attr] for attr in pw_attr_list}

        # Use more attributes that are not taken as they are
        pw_cfg.update({KERNEL_SIZE: (1, 1),
                       STRIDES: (1, 1),
                       PADDING: PAD_VALID,
                       FILTERS: dense_cfg[UNITS]})

        # Create the point-wise layer
        pw_layer = Conv2D(**pw_cfg)
        pw_layer.build(wrapped_layer.layer.input_shape)

        # Create and set the point-wise weights to assign
        dense_kernel = wrapped_layer.layer.kernel
        pw_weights = []
        pw_kernel = np.reshape(wrapped_layer.get_weights()[0],
                               (1, 1, dense_kernel.get_shape()[0], dense_cfg[UNITS]))

        pw_weights.append(pw_kernel)
        if wrapped_layer.layer.use_bias:
            pw_bias = wrapped_layer.get_weights()[2]
            pw_weights.append(pw_bias)

        pw_layer.set_weights(pw_weights)

        # Now that we have the point-wise to replace the dense layer,
        # we need to wrap it using qi.KerasQuantizationWrapper with a new
        # relevant quantizers.
        # Create new kernel quantizer
        pw_kernel_quantizer_cfg = wrapped_layer.weights_quantizers[KERNEL].get_config()

        # In Conv2D channel axis is 3 and not 1 as in Dense
        pw_kernel_quantizer_cfg[keras_inferable_constants.CHANNEL_AXIS] = CONV_KERNEL_CHANNEL_AXIS

        # Unquantized weight to conv layer has 4 dimensions (unlike dense which varies)
        pw_kernel_quantizer_cfg[keras_inferable_constants.INPUT_RANK] = CONV_INPUT_CHANNELS_DIM

        assert isinstance(pw_kernel_quantizer_cfg[keras_inferable_constants.THRESHOLD], np.ndarray), f'Expected to find threshold which is a numpy array, but found: {type(pw_kernel_quantizer_cfg[keras_inferable_constants.THRESHOLD])}'
        pw_kernel_quantizer_cfg[keras_inferable_constants.THRESHOLD] = list(pw_kernel_quantizer_cfg[keras_inferable_constants.THRESHOLD])

        # Now that we have the point-wise quantizer we can instantiate it
        quantizer_class = type(wrapped_layer.weights_quantizers[KERNEL])
        pw_quantizer = quantizer_class(**pw_kernel_quantizer_cfg)
        pw_weights_quantizers = copy.deepcopy(wrapped_layer.weights_quantizers)
        pw_weights_quantizers[KERNEL] = pw_quantizer

        # Wrap pw with the new quantizers (the activation is not affected thus we take the Dense quantizers)
        wrapped_pw = qi.KerasQuantizationWrapper(pw_layer,
                                                 pw_weights_quantizers,
                                                 wrapped_layer.activation_quantizers)

        # Compute the shape that the input to the new layer should be reshaped into
        # Example: Dense kernel with the following shape (3, 20) expects to have input with the
        # next dimensions (BATCH_SIZE, x0, x1, ..., xn, 20).
        # Conv layer expects 4-rank input. Thus, the input is reshaped to (BATCH_SIZE, 1, x0*x1*...*xn, 20)
        dim = wrapped_layer.layer.input_shape[1:-1]
        target_shape = (1, int(np.prod(dim))) + (dense_kernel.get_shape()[0],)

        return Sequential([
            Reshape(target_shape=target_shape),
            wrapped_pw,
            Reshape(wrapped_layer.layer.output_shape[1:])
        ])

    def export(self) -> None:
        """
        Export a fully quantized model to its int8 tflite model.
        """

        def _substitute_model(wrapped_layer: qi.KerasQuantizationWrapper) -> keras.layers.Layer:
            assert self.is_layer_exportable_fn(
                wrapped_layer), f'Layer {wrapped_layer.get_config()} did not pass validation'

            # In order to support dense quantization using per-channel quantization (which is
            # unsupported in TFLITE int models) we substitute each dense layer to its equivalent
            # point-wise convolution.
            if isinstance(wrapped_layer.layer, Dense):
                return self._get_pointwise_layer_to_replace_dense(wrapped_layer)

            return wrapped_layer

        # Transform the model to a new model that can be converted to int8 models.
        # For example: replace dense layers with point-wise layers (to support per-channel quantization)
        self.transformed_model = clone_model(self.model,
                                             clone_function=_substitute_model)

        # Convert model to int8 representation
        converter = tf.lite.TFLiteConverter.from_keras_model(self.transformed_model)
        converter.optimizations = [tf.lite.Optimize.DEFAULT]
        self.exported_model = converter.convert()

        Logger.info(f'Exporting INT8 tflite model to: {self.save_model_path}')
        with open(self.save_model_path, 'wb') as f:
            f.write(self.exported_model)



# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, List, Callable

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.gptq.common.gptq_constants import REG_DEFAULT, REGULARIZATION_VALUES


class GPTQQuantizerConfig:
    """
    A base class to define specific quantizer configuration for GPTQ quantizer.
    """

    def __init__(self):
        self.n_batches = None

    def get_regularization_value(self, fxp_model: Any, **kwargs) -> float:
        """
        Computes a regularization value for the quantizer's loss (if needed).
        In the base class it only returns 0, to be used for GPTQ quantizers that don't require regularization.

        Args:
            fxp_model: The quantized model that is being trained.
            **kwargs: Additional arguments for the quantizer regularization computation.

        Returns: The regularization value.
        """

        return 0

    def set_num_batches(self, num_batches: int):
        """
        Allows to set the number of batches that the quantizer uses for training (in each epoch).

        Args:
            num_batches: number of batches to be set.

        """
        self.n_batches = num_batches


class SoftQuantizerConfig(GPTQQuantizerConfig):
    def __init__(self, entropy_regularization: float = REG_DEFAULT):
        """
        Initializes an object that holds the arguments that are needed for soft rounding quantizer.

        Args:
            entropy_regularization (float): A floating point number that defines the gumbel entropy regularization factor.
        """

        super().__init__()
        self.entropy_regularization = entropy_regularization


    def get_regularization_value(self, fxp_model: Any, **kwargs) -> float:
        """
        Computes a regularization value for the soft quantizer.

        Args:
            fxp_model: The quantized model that is being trained.
            **kwargs: Additional arguments for the quantizer regularization computation.

        Returns: The regularization value.
        """

        soft_rounding_reg_values = kwargs.get(REGULARIZATION_VALUES)

        if soft_rounding_reg_values is None:
            Logger.error("No regularization values has been provided for computing the regularization "  # pragma: no cover
                         "of the soft quantizer.")
        if not isinstance(soft_rounding_reg_values, List):
            Logger.error("The provided regularization values parameter of the soft quantizer "  # pragma: no cover
                         "is not compatible (should be a list).")

        reg = 0

        for sq in soft_rounding_reg_values:
            reg += sq

        return self.entropy_regularization * reg

